from django.contrib.auth import get_user_model
from django.test import TestCase
from django.test.client import RequestFactory
from django.urls.base import reverse
from django.urls.exceptions import NoReverseMatch
from edc_dashboard import url_names

from ...navbar import Navbar
from ...navbar_item import NavbarItem, NavbarItemError
from ...site_navbars import AlreadyRegistered, site_navbars

User = get_user_model()


class TestNavbar(TestCase):
    @classmethod
    def setUpClass(cls):
        url_names.register("dashboard_url", "dashboard_url", "edc_dashboard")
        # GroupPermissionsUpdater(
        #     codenames_by_group=get_default_codenames_by_group(), verbose=True
        # )
        return super().setUpClass()

    def setUp(self):
        site_navbars._registry = {}
        self.user = User.objects.create_superuser("user_login", "u@example.com", "pass")
        rf = RequestFactory()
        self.request = rf.request()
        self.request.user = self.user

    def create_navbar(self):
        testnavbar = Navbar(name="pharmacy_dashboard")
        testnavbar.append_item(
            NavbarItem(
                name="navbar1",
                title="Navbar1",
                label="one",
                codename="edc_navbar.navbar1",
                url_name="navbar_one_url",
            )
        )

        testnavbar.append_item(
            NavbarItem(
                name="navbar2",
                title="Navbar2",
                label="two",
                codename="edc_navbar.navbar2",
                url_name="navbar_two_url",
            )
        )
        return testnavbar

    def test_urls(self):
        reverse("navbar_one_url")
        reverse("navbar_two_url")

    def test_site_navbar_register(self):
        navbar = self.create_navbar()
        site_navbars.register(navbar)
        self.assertTrue(navbar.name in site_navbars.registry)
        self.assertRaises(AlreadyRegistered, site_navbars.register, navbar)

    def test_navbar_item_raises_missing_url(self):
        self.assertRaises(
            NavbarItemError,
            NavbarItem,
            name="navbar_item_one",
            label="Navbar Item One",
            title="navbar_item_one",
            url_name=None,
        )

    def test_navbar_item_raises_bad_url(self):
        self.assertRaises(
            NoReverseMatch,
            NavbarItem,
            name="navbar_item_one",
            label="Navbar Item One",
            title="navbar_item_one",
            url_name="blahblahblah",
        )

    def test_navbar_item_ok(self):
        navbar_item = NavbarItem(
            name="navbar_item_one",
            label="Navbar Item One",
            title="navbar_item_one",
            url_name="navbar_one_url",
            codename="edc_navbar.nav_one",
        )
        self.assertEqual(navbar_item.name, "navbar_item_one")
        self.assertEqual(navbar_item.title, "navbar_item_one")
        self.assertEqual(navbar_item.label, "Navbar Item One")

    def test_render_navbar_item_to_string(self):
        navbar_item_selected = "navbar_item_one"
        navbar_item = NavbarItem(
            name="navbar_item_one",
            label="Navbar Item One",
            title="navbar_item_one",
            url_name="navbar_one_url",
            codename="edc_navbar.nav_one",
        )
        template_string = navbar_item.render(
            request=self.request, navbar_item_selected=navbar_item_selected
        )
        self.assertIn(navbar_item.name, template_string)
        self.assertIn(navbar_item.title, template_string)
        self.assertIn(navbar_item.reversed_url, template_string)
        self.assertIn(navbar_item.label, template_string)

    def test_navbar_repr(self):
        navbar = Navbar(name="default")
        self.assertTrue(repr(navbar))

    def test_render_navbar_item_to_string_fa_icon(self):
        navbar_item_selected = "navbar_item_one"
        navbar_item = NavbarItem(
            name="navbar_item_one",
            label="Navbar Item One",
            title="navbar_item_one",
            url_name="navbar_one_url",
            fa_icon="far fa-user-circle",
            codename="edc_navbar.nav_one",
        )
        template_string = navbar_item.render(
            self.request, navbar_item_selected=navbar_item_selected
        )
        self.assertIn(navbar_item.fa_icon, template_string)

    def test_render_navbar_item_to_string_icon(self):
        navbar_item_selected = "navbar_item_one"
        navbar_item = NavbarItem(
            name="navbar_item_one",
            label="Navbar Item One",
            title="navbar_item_one",
            url_name="navbar_one_url",
            icon="medicine.png",
            codename="edc_navbar.nav_one",
        )
        template_string = navbar_item.render(
            self.request, navbar_item_selected=navbar_item_selected
        )
        self.assertIn(navbar_item.icon, template_string)



from edc_auth.auth_objects import AUDITOR, CLINIC, EVERYONE
from edc_auth.site_auths import site_auths

from .auth_objects import navbar_tuples, remove_permissions_to_edc_navbar_model

site_auths.add_post_update_func(remove_permissions_to_edc_navbar_model)
site_auths.add_custom_permissions_tuples(
    model="edc_navbar.navbar", codename_tuples=navbar_tuples
)
site_auths.update_group(
    "edc_navbar.nav_screening_section", "edc_navbar.nav_subject_section", name=AUDITOR
)
site_auths.update_group(
    "edc_navbar.nav_screening_section", "edc_navbar.nav_subject_section", name=CLINIC
)

site_auths.update_group(
    "edc_navbar.nav_administration",
    "edc_navbar.nav_home",
    "edc_navbar.nav_logout",
    "edc_navbar.nav_public",
    name=EVERYONE,
)



class NavbarError(Exception):
    pass


class Navbar:

    """A class to contain a list of navbar items. See NavbarItem."""

    def __init__(self, name=None, navbar_items=None):
        self.name = name
        self.items = navbar_items or []
        self.rendered_items = []
        self.codenames = {}

    def __repr__(self):
        return f"{self.__class__.__name__}(name={self.name}, items='{self.items}')"

    def __iter__(self):
        return iter(self.items)

    def append_item(self, navbar_item=None):
        self.items.append(navbar_item)
        if not navbar_item.codename:
            raise NavbarError(f"Invalid codename. Got None. See {repr(navbar_item)}.")
        else:
            codename_tuple = (
                navbar_item.codename,
                f'Can access {" ".join(navbar_item.codename.split("_"))}',
            )
            self.codenames.update({navbar_item.codename: codename_tuple})

    def render(self, selected_item=None, request=None, **kwargs):
        """Renders the navbar.

        Note: usually called in NavbarViewMixin.
        """
        self.rendered_items = []
        for item in self.items:
            if item.codename and item.codename not in self.codenames:
                raise NavbarError(
                    f"Permission code is invalid. "
                    f"Expected one of {list(self.codenames.keys())}."
                    f" Got {item.codename}."
                )
            if not item.codename or (item.codename and request.user.has_perm(item.codename)):
                self.rendered_items.append(
                    item.render(selected_item=selected_item, request=request, **kwargs)
                )

    def show_user_permissions(self, user=None):
        """Returns the permissions required to access this Navbar
        and True if the given user has such permissions.
        """
        permissions = {}
        for navbar_item in self.items:
            has_perm = {}
            has_perm.update({navbar_item.codename: user.has_perm(navbar_item.codename)})
            permissions.update({navbar_item.name: has_perm})
        return permissions



navbars = [
    "edc_navbar.nav_administration",
    "edc_navbar.nav_action_item_section",
    "edc_navbar.nav_ae_section",
    "edc_navbar.nav_data_manager_section",
    "edc_navbar.nav_home",
    "edc_navbar.nav_export_section",
    "edc_navbar.nav_lab_aliquot",
    "edc_navbar.nav_lab_manifest",
    "edc_navbar.nav_lab_pack",
    "edc_navbar.nav_lab_process",
    "edc_navbar.nav_lab_receive",
    "edc_navbar.nav_lab_requisition",
    "edc_navbar.nav_lab_section",
    "edc_navbar.nav_logout",
    "edc_navbar.nav_pharmacy_section",
    "edc_navbar.nav_public",
    "edc_navbar.nav_screening_section",
    "edc_navbar.nav_enrolment_section",
    "edc_navbar.nav_subject_section",
    "edc_navbar.nav_tmg_section",
]

navbar_tuples = []
for codename in navbars:
    navbar_tuples.append((codename, f"Can access {codename.split('.')[1]}"))


def remove_permissions_to_edc_navbar_model(auth_updater):
    for group in auth_updater.group_model_cls.objects.all():
        auth_updater.remove_permissions_by_codenames(
            group=group,
            codenames=[
                "edc_navbar.add_navbar",
                "edc_navbar.change_navbar",
                "edc_navbar.delete_navbar",
                "edc_navbar.view_navbar",
            ],
        )



from django.apps import apps as django_apps
from django.views.generic.base import ContextMixin

from .get_default_navbar import get_default_navbar
from .site_navbars import site_navbars


class NavbarViewMixin(ContextMixin):

    navbar_selected_item = None
    navbar_name = get_default_navbar()

    def get_context_data(self, **kwargs):
        """Add rendered navbar <navbar_name> to the context for
        this view.

        Also adds the "default" navbar.
        """
        context = super().get_context_data(**kwargs)
        return self.get_navbar_context_data(context)

    def get_navbar_name(self):
        return self.navbar_name

    def get_navbar_context_data(self, context):
        navbar = site_navbars.get_navbar(name=self.get_navbar_name())
        navbar.render(selected_item=self.navbar_selected_item, request=self.request)
        app_config = django_apps.get_app_config("edc_navbar")
        default_navbar_name = app_config.default_navbar_name

        if default_navbar_name and self.get_navbar_name() != default_navbar_name:
            default_navbar = site_navbars.get_navbar(name=default_navbar_name)
            default_navbar.render(
                selected_item=self.navbar_selected_item, request=self.request
            )
            context.update(
                default_navbar=default_navbar, default_navbar_name=default_navbar_name
            )
        context.update(navbar=navbar)
        return context



from django.conf import settings
from django.urls import path
from django.views.generic.base import View

if settings.APP_NAME == "edc_navbar":

    # needed for tests
    app_name = "edc_navbar"

    urlpatterns = [
        path(r"one/", View.as_view(), name="navbar_one_url"),
        path(r"two/", View.as_view(), name="navbar_two_url"),
        path(r"logout", View.as_view(), name="logout"),
        path(r"administration", View.as_view(), name="administration_url"),
        path(r"", View.as_view(), name="home_url"),
    ]



from django.db import models


class Navbar(models.Model):

    """Dummy model to create content_type to link with permissions.

    Permissions are added against this models content_type.

    See also: edc_auth.permissions_updater
    """

    id = models.BigAutoField(primary_key=True)

    class Meta:
        verbose_name = "Navbar"



import copy

from django.apps import apps as django_apps
from django.conf import settings
from django.core.management.color import color_style
from django.template.loader import render_to_string
from django.urls.base import reverse
from django.urls.exceptions import NoReverseMatch
from edc_constants.constants import IGNORE, WARN
from edc_dashboard.url_names import InvalidUrlName, url_names
from edc_dashboard.utils import get_bootstrap_version

style = color_style()

EDC_NAVBAR_VERIFY_ON_LOAD = getattr(settings, "EDC_NAVBAR_VERIFY_ON_LOAD", None)


class NavbarItemError(Exception):
    pass


class PermissionsCodenameError(Exception):
    pass


class NavbarItem:

    """A class that represents a single item on a navbar."""

    template_name = f"edc_navbar/bootstrap{get_bootstrap_version()}/navbar_item.html"

    def __init__(
        self,
        name=None,
        title=None,
        label=None,
        alt=None,
        url_name=None,
        html_id=None,
        glyphicon=None,
        fa_icon=None,
        icon=None,
        icon_width=None,
        icon_height=None,
        no_url_namespace=None,
        active=None,
        codename=None,
    ):
        self._reversed_url = None
        self._url_name = None
        self.active = active
        self.alt = alt or label or name

        if fa_icon and fa_icon.startswith("fa-"):
            self.fa_icon = f"fa {fa_icon}"
        else:
            self.fa_icon = fa_icon

        self.glyphicon = glyphicon
        self.html_id = html_id or name
        self.icon = icon
        self.icon_height = icon_height
        self.icon_width = icon_width
        self.label = label
        self.name = name
        self.no_url_namespace = no_url_namespace
        self.title = title or self.label or name.title()  # the anchor title

        self.url_name = url_name

        self.check()

        app_label, _codename = self.verify_codename(codename)
        self.codename = f"{app_label}.{_codename}"

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(name={self.name}, "
            f"title={self.title}<url_name={self.url_name}>)"
        )

    def __str__(self):
        return f"{self.name}<url={self.url_name}>"

    def get_context(self, selected_item=None, **kwargs):
        """Returns a dictionary of context data."""
        context = copy.copy(self.__dict__)
        context.update(reversed_url=self.reversed_url, url_name=self.url_name)
        context.update(**kwargs)

        if selected_item == self.name or self.active:
            context.update(active=True)
        return context

    def render(self, request=None, **kwargs):
        """Render to string the template and context data.

        If permission codename is specified, check the user
        has permissions. If not return a disabled control.
        """
        context = self.get_context(**kwargs)
        if not self.codename:
            context.update(has_navbar_item_permission=True)
        else:
            context.update(has_navbar_item_permission=request.user.has_perm(self.codename))
        return render_to_string(template_name=self.template_name, context=context)

    def verify_codename(self, dotted_codename=None):
        if not dotted_codename:
            raise PermissionsCodenameError(
                f"Invalid codename. May not be None. See {repr(self)}"
            )
        try:
            app_label, codename = dotted_codename.split(".")
        except ValueError:
            raise PermissionsCodenameError(
                f"Invalid codename. Got '{dotted_codename}'. See {repr(self)}."
            )
        if app_label not in [a.name for a in django_apps.get_app_configs()]:
            raise PermissionsCodenameError(
                f"Invalid app_label in codename. Expected format "
                f"'<app_label>.<some_codename>'. Got {dotted_codename}. "
                f"See {repr(self)}"
            )
        return app_label, codename

    @property
    def url_name(self):
        return self._url_name

    @url_name.setter
    def url_name(self, value):
        try:
            self._url_name = url_names.get(value)
        except InvalidUrlName:
            self._url_name = value.split(":")[1] if self.no_url_namespace else value
        if not self._url_name:
            raise NavbarItemError(f"'url_name' not specified. See {repr(self)}")

    @property
    def reversed_url(self):
        reversed_url = "#"
        if self.url_name != "#":
            try:
                reversed_url = reverse(self.url_name)
            except NoReverseMatch as e:
                msg = f"{e}. See {repr(self)}."
                if EDC_NAVBAR_VERIFY_ON_LOAD == IGNORE:
                    pass
                elif EDC_NAVBAR_VERIFY_ON_LOAD == WARN:
                    print(style.ERROR(msg))
                else:
                    raise NoReverseMatch(f"{e}. See {repr(self)}.")
        return reversed_url

    def check(self):
        self.reversed_url



from django.urls import include, path
from django.views.generic.base import View

urlpatterns = [
    path(r"edc_data_manager/", include("edc_data_manager.urls")),
    path(r"edc_pharmacy/", include("edc_pharmacy.urls")),
    path(r"edc_lab/", include("edc_lab.urls")),
    path(r"edc_locator/", include("edc_locator.urls")),
    path(r"edc_device/", include("edc_device.urls")),
    path(r"edc_adverse_event/", include("edc_adverse_event.urls")),
    path(r"edc_visit_schedule/", include("edc_visit_schedule.urls")),
    path(r"edc_navbar/", include("edc_navbar.urls")),
    path(r"edc_navbar/", include("edc_consent.urls")),
    path(r"edc_navbar/", include("edc_protocol.urls")),
    path(r"edc_adverse_event/", include("edc_adverse_event.urls")),
    path(r"edc_dashboard/", include("edc_dashboard.urls")),
    path(r"edc_export/", include("edc_export.urls")),
    path(r"", View.as_view(), name="navbar_one_url"),
    path(r"", View.as_view(), name="navbar_two_url"),
    path(r"", View.as_view(), name="logout"),
    path(r"", View.as_view(), name="administration_url"),
    path(r"", View.as_view(), name="home_url"),
]



import sys

from django.apps import AppConfig as DjangoAppConfig
from django.conf import settings
from django.core.management.color import color_style

style = color_style()


class AppConfig(DjangoAppConfig):
    name = "edc_navbar"
    verbose_name = "Edc Navbar"
    register_default_navbar = True
    default_navbar_name = getattr(settings, "DEFAULT_NAVBAR_NAME", "default")

    def ready(self):
        from .site_navbars import site_navbars

        sys.stdout.write(f"Loading {self.verbose_name} ...\n")
        site_navbars.autodiscover()
        sys.stdout.write(f" Done loading {self.verbose_name}.\n")



from django import template
from django.urls import reverse
from django.urls.exceptions import NoReverseMatch
from edc_dashboard.utils import get_bootstrap_version

register = template.Library()


@register.inclusion_tag(
    f"edc_navbar/bootstrap{get_bootstrap_version()}/edc_navbar.html", takes_context=True
)
def show_edc_navbar(context):
    auth_user_change_url = None
    try:
        user = context.get("request").user
    except AttributeError:
        user = None
    else:
        try:
            auth_user_change_url = reverse("admin:auth_user_change", args=(user.id,))
        except NoReverseMatch:
            pass
    return dict(
        auth_user_change_url=auth_user_change_url,
        default_navbar=context.get("default_navbar"),
        navbar=context.get("navbar"),
        user=user,
    )
